import os
import sys
import time


__all__ = ['Logger']

class Params:
    log_colors = {
        'bebug': 'purple',
        'info': 'green',
        'warning': 'yellow',
        'error': 'red',
        'critical': 'bold_red',
        'train': 'cyan',
        'test': 'blue'
    }
    log_colors_code = {
        'purple': '\033[35m{}\033[0m',
        'green':'\033[32m{}\033[0m',
        'yellow': '\033[33m{}\033[0m',
        'red': '\033[31m{}\033[0m',
        'black': '\033[30m{}\033[0m',
        'cyan': '\033[36m{}\033[0m',
        'blue': '\033[34m{}\033[0m'
    }
    log_level = {
        'debug':10,
        'info':20,
        'train':21,
        'test':22,
        'warning':30,
        'error':40,
        'critical':50
    }
    log_level_default = None
    log_name = None
    log_file = ''
    file = None
    write_file_mode = 0
    overwrite=False
    last_msg = ''
    
class Logger():
    def __init__(self, name="", level="INFO", log_file='', write_file_mode=0, overwrite=False, stream='stderr'):
        """Logs are printed on the console and stored in files.
        
        Args:
            name: log subject name.
            level: log printing level.
            log_file: log file path.
            write_file_mode: 1 is fast mode, 0 is slow mode, default 0.
            overwrite: whether overwrite log file.
            stream: stderr or stdout, a file-like object (stream), defaults to the current sys.stderr.
        """
        self.params = Params()
        self.params.log_name = 'root' if name=="" else name
        if str.lower(level) in self.params.log_level:
            self.params.log_level_default = str.lower(level)
        else:
            raise ValueError("`level` must in param dict `Logger.params.log_level`. ")
        
        self.params.write_file_mode = write_file_mode
        self.params.overwrite = overwrite
        self.params.stream = stream
        self.params.time = time.time()
        self.params.time_start = self.params.time
        self.update_log_file(log_file)
            
    def log(self, level, msg, write_file, enter, close):
        """Logs are printed on the console and stored in files.
        
        Args:
            level: log printing level.
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
        """
        if self.params.log_level[str.lower(level)]<self.params.log_level[self.params.log_level_default]:
            return 
        
        time_now = time.time()
        end = time_now-self.params.time_start if close else time_now-self.params.time
        self.params.time = time_now
        if end < 60:
            msg = f'[{end:.2f} sec]: ' + msg
        elif end < 3600:
            msg = "[%d min %.2f s]: " % divmod(end, 60) + msg
        else:
            msg = f"[{end // 3600:.0f} hour %d min %.0f s]: " % divmod(end % 3600, 60) + msg
        msg = (f"[{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time_now))}]" +
               f" [{self.params.log_name}] [{str.upper(level)}] " + msg)
        if write_file:
            self.write(msg)

        if len(self.params.last_msg)>len(msg):
            msg = msg+' '*(len(self.params.last_msg)-len(msg))
        self.params.last_msg = msg
        try:
            msg = self.params.log_colors_code[self.params.log_colors[str.lower(level)]].format(msg)
        except:
            msg = f'\033[30m{msg}\033[0m'
        print(msg, end='\n' if enter else '\r', file=sys.stderr if self.params.stream=='stderr' else sys.stdout)
        
        if close:
            self.close()
        
    def write(self, msg):
        """Logs write to the log file.
        
        Args:
            msg: log message.
        """
        try:
            self.params.file.write(msg + '\n')
        except:
            try:
                with open(self.params.log_file, 'a+') as f:
                    f.write(msg + '\n')
            except:
                pass
    
    def debug(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("DEBUG", msg, write_file, enter, close)

    def info(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("INFO", msg, write_file, enter, close)

    def warning(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("WARNING", msg, write_file, enter, close)

    def error(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("ERROR", msg, write_file, enter, close)

    def critical(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("CRITICAL", msg, write_file, enter, close)
        
    def train(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("TRAIN", msg, write_file, enter, close)
    
    def test(self, msg, write_file=False, enter=True, close=False):
        """Logs are printed on the console and stored in files.
        
        Args:
            msg: log printing message.
            write_file: whether to control log write to the log file.
            enter: whether to wrap print.
            close: whether to end the entire log notification.
        """
        self.log("TEST", msg, write_file, enter, close)

    def update_log_file(self, log_file):
        """Update log file path.
        
        Args:
            log_file: log file path.
        """
        real_path = os.path.realpath(os.path.dirname(log_file))
        if log_file=='':
            pass
        elif real_path != os.path.abspath(log_file):
            if not os.path.exists(real_path):
                os.makedirs(real_path)
            self.params.log_file = log_file
            if self.params.write_file_mode:
                self.close()
                mode = 'w' if self.params.overwrite else 'a+'
                self.params.file = open(self.params.log_file, mode)
        else:
            raise ValueError("`log_file` must is file name.")
    
    def close(self):
        """Close log ssesion."""
        if self.params.file is not None:
            self.params.file.close()



from scipy.stats import norm


__all__ = ['numeric_binarizer', 'numeric_bucketized', 'numeric_padding', 'numeric_outlier']

def numeric_binarizer(feature, feature_scale=None):
    """Encode labels with value between 0 and 1.
    
    Args:
        feature: pd.Series, sample feature.
        feature_scale: float, feature.mean().
    Returns:
        normalize feature and feature_scale.
    """
    scale = feature_scale if feature_scale is not None else feature.mean()
    t = feature.clip(upper=scale).replace({scale:scale+0.1}).clip(scale).replace({scale:0, scale+0.1:1}).astype('int')
    return t, scale

def numeric_bucketized(feature, boundaries, miss_pad=-1, score=None, miss_score=None, dtype='int64', mode=1):
    """feature bucket.
    
    if mode is True:
        Buckets include the right boundary, and exclude the left boundary. 
        Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.], (0., 1.], (1., 2.], and (2., +inf).
    else:
        Buckets include the left boundary, and exclude the right boundary. 
        Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.), [0., 1.), [1., 2.), and [2., +inf).
        
    Args:
        feature: pd.Series, sample feature.
        boundaries: A sorted list or tuple of floats specifying the boundaries.
        miss_pad: default -1, feature fillna value.
        dtype: default 'int64', return transfrom dtypes.
        score: None, A score list or tuple of floats specifying the boundaries.
        miss_score: None, score fillna value.
        mode: True.
    Returns:
        normalize feature.
    """
    t = feature.copy()
    bound = sorted(boundaries)
    if mode:
        for i in range(len(bound)):
            if i==0:
                t[feature<=bound[i]] = i
            else:
                t[(feature>bound[i-1])&(feature<=bound[i])] = i
        t[feature>bound[i]] = i+1
    else:
        t[feature<bound[0]] = 0
        for r, i in enumerate(bound):
            t[feature>=i] = r+1
    t = t.fillna(miss_pad).astype(dtype)
    if isinstance(score, (tuple, list)):
        t = t.replace({i:j for i,j in enumerate(score)})
        if miss_score is not None:
            t = t.replace({miss_pad:miss_score}) 
    return t

def numeric_padding(feature, method='mean', feature_scale=None):
    """feature fillna method.
    
    Args:
        feature: pd.Series, sample feature.
        method: default 'mean', one of 'mean' or 'median'.
        feature_scale: float, feature.mean() or feature.median().
    Returns:
        normalize feature and feature_scale.
    """
    assert method in ['mean', 'median'], "`method` should be one of ['mean', 'median']."
    scale_dict = {'mean':feature.mean(), 'median':feature.median()}
    scale = feature_scale if feature_scale is not None else scale_dict[method]
    t = feature.fillna(scale)
    return t, scale

def numeric_outlier(feature, keep_rate=0.9545, mode='right', feature_scale=None):
    """feature clip outlier.
    
    Args:
        feature: pd.Series, sample feature.
        keep_rate: default 0.9545, 
        method: default 'right', one of ['left', 'right', 'both'], statistical distribution boundary.
        feature_scale: list or tuple, [feature.mean(), feature.std()].
    Returns:
        normalize feature and feature_scale.
    """
    assert mode in ['left', 'right', 'both'], "`mode` should be one of ['left', 'right', 'both']."
    scale = feature_scale if feature_scale is not None else (feature.mean(), feature.std())
    if mode=='both':
        clip_dict = (feature.mean()+norm.ppf((1-keep_rate)/2)*feature.std(), feature.mean()+norm.ppf(keep_rate+(1-keep_rate)/2)*feature.std())
    elif mode=='right':
        clip_dict = (feature.min(), feature.mean()+norm.ppf(keep_rate)*feature.std())
    else:
        clip_dict = (feature.mean()+norm.ppf(1-keep_rate)*feature.std(), feature.max())
    t = feature.clip(clip_dict[0], clip_dict[1])
    return t, scale



import pandas as pd
import sqlalchemy as sa

__all__ = ['HiveEngine']

class HiveTableTypes():
    BigInteger = sa.types.BigInteger
    Boolean = sa.types.Boolean
    DATE = sa.types.DATE
    LargeBinary = sa.types.LargeBinary
    String = sa.types.String
    

class HiveEngine():
    def __init__(self, cur):
        self.engine = sa.create_engine(cur)
        self.tpyes = HiveTableTypes
    
    def create_table(self, table, feature, sep=','):
        t = ', '.join([i[0]+' '+i[1]+" comment '"+i[2]+"'" for i in feature])
        sql = f"CREATE TABLE IF NOT EXISTS {table} ({t}) row format delimited fields terminated by '{sep}'"
        return pd.io.sql.execute(sql, self.engine)
    
    def drop_table(self, table):
        sql = f"DROP TABLE IF EXISTS {table}"
        return pd.io.sql.execute(sql, self.engine)
    
    def alter_table_comment(self, table, comment='', property_name=''):
        sql = f"ALTER TABLE {table} SET TBLPROPERTIES('comment'='{comment}', 'property_name'='{property_name}')"
        return pd.io.sql.execute(sql, self.engine)
    
    def alter_table_delimiter(self, table, delimiter):
        sql = f"ALTER TABLE {table} SET SERDEPROPERTIES('field.delim'='{delimiter}')"
        return pd.io.sql.execute(sql, self.engine)
    
    def alter_table_name(self, old_table, new_table):
        sql = f"ALTER TABLE {old_table} RENAME TO {new_table}"
        return pd.io.sql.execute(sql, self.engine)
    
    def alter_column(self, table, old_column, new_column, dtype, comment):
        sql = f"ALTER TABLE {table} CHANGE COLUMN {old_column} {new_column} {dtype} COMMENT '{comment}'"
        return pd.io.sql.execute(sql, self.engine)



import numpy as np
import pandas as pd

__all__ = ['euclidean', 'manhattan', 'chebyshev', 'minkowski', 'hamming',
           'jaccard', 'pearson', 'cosine', 'levenshtein'
          ]

def euclidean(x, y, normalize=False):
    """euclidean distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
        normalize: default False, std=pd.concat([x, y]).std() if normalize else 1.
    Returns:
        euclidean distance value.
    """
    std = pd.concat([x, y]).std() if normalize else 1
    return np.sqrt(np.sum(np.square((x-y)/std)))

def manhattan(x, y):
    """manhattan distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        manhattan distance value.
    """
    return np.sum(np.abs(x-y))

def chebyshev(x, y):
    """chebyshev distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        chebyshev distance value.
    """
    return np.max(x-y)

def minkowski(x, y, p):
    """minkowski distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
        p: int, norm dimension.
    Returns:
        minkowski distance value.
    """
    return np.sqrt(np.sum(np.power(np.abs(x-y), p)))

def hamming(x, y):
    """hamming distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        hamming distance value.
    """
    return len(x)-np.sum(np.equal(x, y))

def jaccard(x, y):
    """jaccard distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        jaccard distance value.
    """
    return 1 - len(set(x).intersection(set(y)))/len(set(x).union(set(y)))

def pearson(x, y):
    """pearson distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        pearson distance value.
    """
    x_mean, y_mean = np.mean(x), np.mean(y)
    d = np.sum((x-x_mean)*(y-y_mean))
    d2 = np.sqrt(np.sum(np.square(x-x_mean)))*np.sqrt(np.sum(np.square(y-y_mean)))
    return 1-d/d2

def cosine(x, y):
    """cosine distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
    Returns:
        cosine distance value.
    """
    return (x*y).sum()/np.sqrt(np.square(x).sum())/np.sqrt(np.square(y).sum())

def levenshtein(x, y, normalize=False):
    """levenshtein distance.
    
    Args:
        x: pd.Series, sample feature value.
        y: pd.Series, sample feature value.
        normalize: default False, if normalize is True, levenshtein distance should be distance/max(len(x), len(y)).
    Returns:
        levenshtein distance value.
    """
    def levenshtein1(s1, s2, normalize=False):
        if len(s1) < len(s2):
            return levenshtein1(s2, s1, normalize)
        if not s2:
            return len(s1)

        a = range(len(s2) + 1)
        for i, c1 in enumerate(s1):
            b = [i+1]
            for j, c2 in enumerate(s2):
                b.append(min(a[j+1]+1, b[j]+1, a[j]+(c1 != c2)))
            a = b

        if normalize:
            return (b[-1] / len(s1))
        return b[-1]
    return pd.DataFrame({'label1':x, 'label2':y}).apply(lambda x:levenshtein1(x[0], x[1], normalize=normalize), axis=1).sum()



from linora.image._image_rescale import *

class ImageRescaleAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def normalize_global(self, mean=None, std=None):
        """Normalize scales `image` to have mean and variance.
    
        This op computes `(x - mean) / std`.
        Args:
            mean: if None, computes image mean.
                  if int or float, customize image all channels mean.
            std: if None, computes image std.
                 if int or float, customize image all channels std.
       Returns:
         The standardized image with same shape as `image`.
       Raises:
         ValueError: if the shape of 'image' is incompatible with this function.
       """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = normalize_global(self.image, mean, std)
        return self
    
    def normalize_channel(self, mean=None, std=None):
        """Normalize scales `image` to have mean and variance.
    
        This op computes `(x - mean) / std`.
        Args:
            mean: if None, computes image mean.
                  if tuple or list, customize image each channels mean,
                  shape should 3 dims.
            std: if None, computes image std.
                 if tuple or list, customize image each channels std,
                 shape should 3 dims.
      Returns:
        The standardized image with same shape as `image`.
      Raises:
        ValueError: if the shape of 'image' is incompatible with this function.
      """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = normalize_channel(self.image, mean, std)
        return self

    def rescale(self, scale):
        """Rescale apply to image.

        new pixel = image * scale
        Args:
            scale: if int float, value multiply with image.
                   if tuple list, randomly picked in the interval
                   `[central_rate[0], central_rate[1])`, value multiply with image.
        Returns:
            The image with same shape as `image`.
        Raises:
            scale type error.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = rescale(self.image, scale)
        return self
    
    
    



import io
import base64

from PIL import Image

__all__ = ['read_image', 'save_image', 'encode_base64', 'decode_base64']

def read_image(filename):
    """
    Args:
        filename: str, image absolute path.
    Returns:
        a Image instance.
    """
    with open(filename, 'rb') as f:
        image = Image.open(io.BytesIO(f.read()))
    return image

def save_image(filename, image, file_format=None, **kwargs):
    """Saves an image stored as a Numpy array to a path or file object.
    Args
        filename: Path or file object.
        image: A PIL Image instance.
        file_format: Optional file format override. If omitted, the
            format to use is determined from the filename extension.
            If a file object was used instead of a filename, this
            parameter should always be used.
        **kwargs: Additional keyword arguments passed to `PIL.Image.save()`.
    """
    if image.mode == 'RGBA' and file_format in ['jpg', 'jpeg']:
#         warnings.warn('The JPG format does not support RGBA images, converting to RGB.')
        image = image.convert('RGB')
    image.save(filename, format=file_format, **kwargs)

def encode_base64(file):
    """encode image to string.
    Args
        file: image file path.
    """
    with open(file, "rb")as f:
        bs64 = base64.b64encode(f.read()).decode()
    return bs64

def decode_base64(file, image_str):
    """decode image to file.
    Args
        file: image file path.
        image_str: image bites string.
    """
    with open(file, "wb") as f:
        f.write(base64.b64decode(image_str))



from linora.image._image_noise import *

class ImageNoiseAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def noise_gaussian(self, scale=1, mean=0.0, std=1.0):
        """Gaussian noise apply to image.

        new pixel = image + gaussian_noise * scale
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth].
            scale: if int or float, value multiply with poisson_noise.
                   if tuple or list, randomly picked in the interval
                   `[scale[0], scale[1])`, value multiply with poisson_noise.
            mean: if int or float, value is gaussian distribution mean.
                  if tuple or list, randomly picked in the interval
                  `[mean[0], mean[1])`, value is gaussian distribution mean.
            std: if int or float, value is gaussian distribution std.
                 if tuple or list, randomly picked in the interval
                 `[std[0], std[1])`, value is gaussian distribution std.
        Returns:
            3-D float Tensor, as per the input.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = noise_gaussian(self.image, scale, mean, std)
        return self
    
    def noise_poisson(self, scale=1, lam=1.0):
        """Poisson noise apply to image.

        new pixel = image + poisson_noise * scale
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth].
            scale: if int or float, value multiply with poisson_noise.
                   if tuple or list, randomly picked in the interval
                   `[scale[0], scale[1])`, value multiply with poisson_noise.
            lam: if int or float, value is poisson distribution lambda.
                 if tuple or list, randomly picked in the interval
                 `[lam[0], lam[1])`, value is poisson distribution lambda.
        Returns:
            3-D float Tensor, as per the input.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = noise_poisson(self.image, scale, lam)
        return self
    
    def noise_mask(self, noise_prob=0.2):
        """Mask noise apply to image.

        With probability `drop_prob`, outputs the input element scaled up by
        `1`, otherwise outputs `0`. 

        Tips:
            1 mean pixel have no change.
            a suitable interval is (0., 0.1].
        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth].
            noise_prob: should be in the interval (0, 1.].
                       if float, the probability that each element is drop.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is drop.
        Returns:
            3-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1.]`.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = noise_mask(self.image, noise_prob)
        return self
    
    def noise_saltpepper(self, noise_prob=0.2):
        """Salt-Pepper noise apply to image.

        The salt-pepper noise is based on the signal-to-noise ratio of the image,
        randomly generating the pixel positions in some images all channel,
        and randomly assigning these pixels to 0 or 255.

        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth].
            noise_prob: should be in the interval (0, 1].
                       if int or float, the probability that each element is kept.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
        Returns:
            3-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1]`.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = noise_saltpepper(self.image, noise_prob)
        return self
    
    def noise_rainbow(image, noise_prob=0.2):
        """Rainbowr noise apply to image.

        The rainbow noise is based on the signal-to-noise ratio of the image,
        randomly generating the pixel positions in some images,
        and randomly assigning these pixels to 0 or 255.

        Args:
            image: Either a 3-D float Tensor of shape [height, width, depth].
            noise_prob: should be in the interval (0, 1].
                       if int or float, the probability that each element is kept.
                       if tuple or list, randomly picked in the interval
                       `[keep_prob[0], keep_prob[1])`, the probability that each element is kept.
        Returns:
            3-D float Tensor, as per the input.
        Raises:
            ValueError: If `keep_prob` is not in `(0, 1]`.
        """
        if type(self.image)!=np.ndarray:
            self.image_to_array(self.image)
        self.image = noise_rainbow(self.image, noise_prob)
        return self
    
    
    
    



import numpy as np
from PIL import ImageEnhance

__all__ = ['enhance_color', 'enhance_contrast', 'enhance_brightness', 'enhance_sharpness']


def enhance_color(image, delta):
    """Adjust image color balance.

    This class can be used to adjust the colour balance of an image, 
    in a manner similar to the controls on a colour TV set. 
    An enhancement factor of 0.0 gives a black and white image. 
    A factor of 1.0 gives the original image.
    
    Args:
    image: a Image instance.
    delta: A floating point value controlling the enhancement. 
           delta 1.0 always returns a copy of the original image, 
           lower factors mean less color, 
           and higher values more. There are no restrictions on this value.
           if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` , value is float multiplier for adjusting color.
    Returns:
        a Image instance.
    """
    if isinstance(delta, (list, tuple)):
        assert delta[0]<delta[1], 'delta should be delta[1] > delta[0].'
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Color(image).enhance(delta)

def enhance_contrast(image, delta):
    """Adjust contrast of RGB or grayscale images.
  
    Contrast is adjusted independently for each channel of each image.
    
    For each channel, this Ops computes the mean of the image pixels in the
    channel and then adjusts each component `x` of each pixel to
    `(x - mean) * delta + mean`.
    
    Tips:
        1 means pixel value no change.
        0 means all pixel equal. 
        a suitable interval is (0, 4].
    Args:
        images: a Image instance.
        delta: if int, float, a float multiplier for adjusting contrast.
               if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` , value is float multiplier for adjusting contrast.
    Returns:
        a Image instance.
    """
    if isinstance(delta, (list, tuple)):
        assert delta[0]<delta[1], 'delta should be delta[1] > delta[0].'
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Contrast(image).enhance(delta)

def enhance_brightness(image, delta):
    """Adjust the brightness of RGB or Grayscale images.
    
    Tips:
        delta extreme value in the interval [-1, 1], >1 to white, <-1 to black.
        a suitable interval is [-0.5, 0.5].
        0 means pixel value no change.
    Args:
        image: Tensor or array. An image.
        delta: if int, float, Amount to add to the pixel values.
               if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` to add to the pixel values.
    Returns:
        a Image instance.
    """
    if isinstance(delta, (list, tuple)):
        assert delta[0]<delta[1], 'delta should be delta[1] > delta[0]'
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Brightness(image).enhance(delta)

def enhance_sharpness(image, delta):
    """Adjust image sharpness.

    This class can be used to adjust the sharpness of an image. 
    An enhancement factor of 0.0 gives a blurred image, 
    a factor of 1.0 gives the original image, 
    and a factor of 2.0 gives a sharpened image.
    
    Args:
    image: a Image instance.
    delta: A floating point value controlling the enhancement. 
           delta 1.0 always returns a copy of the original image, 
           lower factors mean less sharpness, 
           and higher values more. There are no restrictions on this value.
           if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` , value is float multiplier for adjusting color.
    Returns:
        a Image instance.
    """
    if isinstance(delta, (list, tuple)):
        assert delta[0]<delta[1], 'delta should be delta[1] > delta[0]'
        delta = np.random.uniform(delta[0], delta[1])
    return ImageEnhance.Sharpness(image).enhance(delta)


def RandomGamma(image, gamma, seed=None, **kwarg):
    """Performs Gamma Correction on the input image.
    
    Also known as Power Law Transform. This function transforms the
    input image pixelwise according to the equation `Out = In**gamma`
    after scaling each pixel to the range 0 to 1.
    
    Tips:
        For gamma greater than 1, the histogram will shift towards left and
        the output image will be darker than the input image.
        For gamma less than 1, the histogram will shift towards right and
        the output image will be brighter than the input image.
        if gamma is 1, image pixel value no change.
    Args:
        image : A Tensor.
        gamma : if int, float, Non negative real number.
                if list, tuple, randomly picked in the interval
                `[delta[0], delta[1])` , value is Non negative real number.
        seed: A Python integer. Used to create a random seed. See
              `tf.set_random_seed` for behavior.
    Returns:
        A float Tensor. Gamma corrected output image.
    Raises:
        ValueError: If gamma is negative.
    References:
        [1] http://en.wikipedia.org/wiki/Gamma_correction
    """
    image = tf.cast(image, dtype=tf.float32)
    if isinstance(gamma, (int, float)):
        assert 0<gamma, 'gamma should be > 0.'
        image = tf.image.adjust_gamma(image, gamma, gain=1)
    elif isinstance(gamma, (list, tuple)):
        assert 0<gamma[0]<gamma[1], 'gamma should be gamma[1] > gamma[0] > 0.'
        random_gamma = tf.random.uniform([], gamma[0], gamma[1], seed=seed)
        image = tf.image.adjust_gamma(image, random_gamma, gain=1)
    else:
        raise ValueError('gamma should be one of int, float, list, tuple.')
    return image if kwarg else image.numpy()

def RandomPencilSketch(image, delta=0.1, seed=None, **kwarg):
    """Adjust the pencil sketch of RGB.
    
    Args:
        image: Tensor or array. An image.
        delta: if int, float, Amount to add to the pixel values.
               if list, tuple, randomly picked in the interval
               `[delta[0], delta[1])` to add to the pixel values.
               a suitable interval is [0.1, 0.5].
        seed: A Python integer. Used to create a random seed. See
             `tf.set_random_seed` for behavior.
    Returns:
        A pencil_sketch-adjusted tensor of the same shape as `image`.
    """
    if isinstance(delta, (list, tuple)):
        random_delta = tf.random.uniform([], delta[0], delta[1], seed=seed)
    else:
        random_delta = delta
        
    temp = tf.cast(image, tf.float32)
    t = temp[:, 1:]-temp[:, :-1]
    grad_y = tf.concat([t[:,0:1], (t[:, 1:]+t[:, :-1])/2, t[:, -1:]], 1)*random_delta
    t = temp[1:, :]-temp[:-1, :]
    grad_x = tf.concat([t[0:1,:], (t[1:, :]+t[:-1, :])/2, t[-1:, :]], 0)*random_delta

    A = tf.math.sqrt(tf.math.square(grad_x)+tf.math.square(grad_y)+1.)
    uni_x = tf.math.divide(grad_x, A)
    uni_y = tf.math.divide(grad_y, A)
    uni_z = tf.math.divide(1., A)

    dx = tf.math.cos(3.141592653589793/2.2)*tf.math.cos(3.141592653589793/4)
    dy = tf.math.cos(3.141592653589793/2.2)*tf.math.sin(3.141592653589793/4)
    dz = tf.math.sin(3.141592653589793/2.2)

    b = tf.clip_by_value(255.*(dx*uni_x + dy*uni_y + dz*uni_z), 0., 255.)
    return b if kwarg else b.numpy()
